/// @file
////////////////////////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2018/19      Christian Lessig, Otto-von-Guericke Universitaet Magdeburg
///
////////////////////////////////////////////////////////////////////////////////////////////////////
///
///  module     : lecture 10
///
///  author     : lessig@isg.cs.uni-magdeburg.de
///
///  project    : GPU Programming
///
///  description: simple Cuda demo
///
////////////////////////////////////////////////////////////////////////////////////////////////////

// includes, system
#include <iostream>
#include <chrono>

// includes, project
#include "cuda_util.h"

////////////////////////////////////////////////////////////////////////////////////////////////////
// transpose of matrix: using only global memory
////////////////////////////////////////////////////////////////////////////////////////////////////
__global__ void
transpose1( const float* a, float* b, int n) {

  const unsigned int j = blockIdx.x * blockDim.x + threadIdx.x;
  const unsigned int i = blockIdx.y * blockDim.y + threadIdx.y;

  if( i < n && j < n) {
    b[j*n + i] = a[i*n + j];
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// transpose of matrix: using shared memory
////////////////////////////////////////////////////////////////////////////////////////////////////
__global__ void
transpose2( const float* a, float* b, int n) {

  __shared__ float smem[1024];

  int j = blockIdx.x * blockDim.x + threadIdx.x;
  int i = blockIdx.y * blockDim.y + threadIdx.y;

  if( i < n && j < n) {
    smem[threadIdx.x * blockDim.x + threadIdx.y] = a[i*n + j];
  }
  __syncthreads();

  j = blockIdx.y * blockDim.x + threadIdx.x;
  i = blockIdx.x * blockDim.y + threadIdx.y;

  if( i < n && j < n) {
    b[i*n + j] = smem[threadIdx.y * blockDim.x + threadIdx.x];
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// initialize Cuda device
////////////////////////////////////////////////////////////////////////////////////////////////////
void
initDevice( int& device_handle) {

  int deviceCount = 0;
  cudaGetDeviceCount(&deviceCount);

  if( 0 == deviceCount) {
    std::cerr << "initDevice() : No CUDA device found." << std::endl;
  }

  // one could implement more complex logic here to find the fastest device
  if( deviceCount > 1) {
    std::cerr << "initDevice() : Multiple CUDA devices found. Using first one." << std::endl;
  }

  // set the device
  cudaSetDevice( device_handle);
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// initialize device memory
////////////////////////////////////////////////////////////////////////////////////////////////////
void
initDeviceMemory( const float* a, const unsigned int n,
                  float*& a_device, float*& b_device)
{
  // allocate device memory
  cudaMalloc((void **) &a_device, sizeof(float) * n * n);
  cudaMalloc((void **) &b_device, sizeof(float) * n * n);

  // copy device memory
  cudaMemcpy( a_device, a, sizeof(float) * n * n, cudaMemcpyHostToDevice );
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// initialize device memory
////////////////////////////////////////////////////////////////////////////////////////////////////
void
getResultDevice( const float* b_device, float* b, const unsigned int n) {

  cudaMemcpy( b, b_device, sizeof(float) * n * n, cudaMemcpyDeviceToHost );
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// free device memory
////////////////////////////////////////////////////////////////////////////////////////////////////
void
freeDeviceMemory( float*& a_device, float*& b_device) {

  cudaFree( a_device);
  cudaFree( b_device);
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// free device memory
////////////////////////////////////////////////////////////////////////////////////////////////////
void
transposeHost( float* a, float* ref, const int n) {

  for( int i = 0; i < n; ++i) {
    for( int j = 0; j < n; ++j) {
      ref[j*n+i] = a[i*n+j];
    }
  }

}

////////////////////////////////////////////////////////////////////////////////////////////////////
// program entry point
////////////////////////////////////////////////////////////////////////////////////////////////////
int
main( int /*argc*/, char** /*argv*/ ) {

  // initialize device
  int device_handle = 0;
  initDevice( device_handle);

  // set up host memory
  const int n = 4096;
  float* a = (float*) malloc( sizeof(float) * n * n);
  float* b = (float*) malloc( sizeof(float) * n * n);
  for( unsigned int i = 0; i < n*n; ++i) {
    a[i] = (float) i;
  }

  float* ref = (float*) malloc( sizeof(float) * n * n);
  transposeHost( a, ref, n);

  // initialize device memory
  float* a_device = nullptr;
  float* b_device = nullptr;
  initDeviceMemory( a, n, a_device, b_device);

  // clean host memory that is no longer needed
  free( a);

  // determine thread layout
  dim3 num_threads_per_block( 32, 32);
  dim3 num_blocks( n / num_threads_per_block.x);
  if( 0 != n % num_threads_per_block.x) {
    num_blocks.x++;
  }
  num_blocks.y = num_blocks.x;

  std::cout << num_threads_per_block.x << " / " << num_threads_per_block.y << std::endl;
  std::cout << num_blocks.x << " / " << num_blocks.y << std::endl;

  auto start = std::chrono::steady_clock::now();

  // run kernel
  for( int k = 0; k < 100; ++k) {
    transpose1<<< num_blocks , num_threads_per_block >>>( a_device, b_device, n);
  }
  cudaDeviceSynchronize();

  typedef std::chrono::milliseconds TimeT;
  auto t = std::chrono::duration_cast<TimeT>(std::chrono::steady_clock::now()-start).count();
  std::cout << "time = " << t << " ms" << std::endl;

  // copy result back to host
  getResultDevice( b_device, b, n);

  // check result
  float err = 0.0;
  for( int i = 0; i < n*n; ++i) {
    err = std::max( err, std::abs(ref[i] - b[i]));
  }
  std::cout << "max err = " << err << std::endl;

#if 0
  for( int i = 0; i < 100; ++i) {
    std::cout << b[i] << " / " << ref[i] << std::endl;
  }
#endif

  // clean up device memory
  freeDeviceMemory( a_device, b_device);

  // clean up host memory
  free( ref);

  return EXIT_SUCCESS;
}
